// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Abstract syntax.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

// System headers.

#ifndef __CSTDIO__
#include <cstdio>
#endif
#ifndef __CSTRING__
#include <cstring>
#endif
#ifndef __SGI_STL_VECTOR
#include <vector>
#endif
#ifndef __SGI_STL_STACK
#include <stack>
#endif
#ifndef _PROPERTY_INFO_H
#include <PropertyInfo.h>
#endif
#ifndef _ENTRY_H
#include <Entry.h>
#endif
#ifndef _ERRORS_H
#include <Errors.h>
#endif
#ifndef	_CONTROL_H
#include <Control.h>
#endif

// Local headers.

#include "AbstractSyntax.h"
#include "ScriptingElement.h"
#include "ScriptingTargets.h"
#include "ScriptingUtils.h"

// GLOBAL DATA:
// Verbosity level: 0 = silent (default), 1 = display commands, 2 = display messages.

extern int* gVerbosity;

// A custom messenger, capable of resolving itself when it is used (if need be).

DMessenger::DMessenger(const char* identifier)
{
	fMessenger = new BMessenger(identifier);
	fIdentifier = identifier;
}

DMessenger::DMessenger(DMessenger* messenger)
{
	fMessenger = new BMessenger(*(messenger->fMessenger));
	fIdentifier = messenger->fIdentifier;
}

DMessenger::DMessenger(BMessenger* messenger)
{
	fMessenger = new BMessenger(*messenger);
	fIdentifier = "";
}

DMessenger::~DMessenger()
{
	delete fMessenger;
}

status_t DMessenger::SendMessage(BMessage* msg, BMessage* reply)
{
	// Try to reconstruct messenger if it was invalid at construction time.
	// This can be attempted only for messengers constructed from MIME strings.
	if (fMessenger && !fMessenger->IsValid())
	{
		if (fIdentifier.size() >= 1)
		{
			delete fMessenger;
			fMessenger = DMessenger::ResolveTarget(fIdentifier.c_str());
		}
	}
	if (fMessenger && fMessenger->IsValid())
		return fMessenger->SendMessage(msg, reply);
	else
		return B_BAD_PORT_ID;
}

bool DMessenger::IsValid()
{
	if (!fMessenger || (fMessenger && !fMessenger->IsValid()))
	{
		delete fMessenger;
		fMessenger = DMessenger::ResolveTarget(fIdentifier.c_str());
	}
	return fMessenger ? fMessenger->IsValid() : false;
}

// Resolve a target.

BMessenger* DMessenger::ResolveTarget(const char* str)
{
	// Ask the roster to resolve a target based on a name or MIME string.
	// Retrieve current list of teams and locate the desired application.
	// Create a messenger for that application.
	BMessenger* msgr = 0;
	BList team_list;
	team_id teamid;
	app_info appinfo;
	be_roster->GetAppList(&team_list);
	int32 items = team_list.CountItems();
	for (int32 i = 0; !msgr && i < items; ++i)
	{
		teamid = (team_id)team_list.ItemAt(i);
		be_roster->GetRunningAppInfo(teamid, &appinfo);
		if (strcmp(appinfo.ref.name, str) == 0)
			msgr = new BMessenger(0, teamid);
		else if (strcmp(appinfo.signature, str) == 0)
			msgr = new BMessenger(appinfo.signature);
	}
	return msgr;
}

// Abstract syntax:
// A generic scripting command.

// The default constructor ensures that the object won't be used.
Command::Command()
:	fMessenger(0), fMessage(B_MESSAGE_TYPE), fReply(B_NO_REPLY), fValue(B_MESSAGE_TYPE),
	fParent(0), fDynamic(0), fSent(false), fInited(false) {}

// Regular constructor.
Command::Command(DMessenger* msgr, const uint32 command)
:	fMessenger(msgr), fMessage(command), fReply(B_REPLY), fValue(B_MESSAGE_TYPE),
	fParent(0), fDynamic(0), fSent(false), fInited(true) {}

// Destructor is currently unused.
Command::~Command() {}

// Run command.
status_t Command::Do()
{
	status_t result = Execute();
	if (result != B_OK)
		result = AutoResolve(result);
	return result;
}

// Run command and display result.
status_t Command::DoAndPrintReply()
{
	status_t result = Execute();
	if (result != B_OK)
		result = AutoResolve(result);
	if (result != B_OK)
		return result;
	PrintResult();
	return B_OK;
}

// Display the command's raw messages.
void Command::PrintContents(bool printReply = true)
{
	printf("  Send: ");
	print_message(&fMessage);
	if (printReply && fSent)
	{
		printf("Result: ");
		PrintResult();
	}
}

// Display command text with correct indentation.
void Command::Print(int indent = 0)
{
	PrintIndentation(indent);
	printf("%s\n", Text().c_str());
}

// Display current indentation.
void Command::PrintIndentation(int indent = 0)
{
	for (int index = 0; index < indent; ++index)
		printf("    ");
}

// Display command result.
void Command::PrintResult()
{
	// Safety check.
	if (fReply.what != B_REPLY)
	{
		puts("No valid reply!");
		return;
	}
	
	// Locate "result" field.
	type_code typefound;
	int32 count;
	fReply.GetInfo(SF_RESULT, &typefound, &count);
	if (!count)
	{
		puts("Empty reply!");
		return;
	}
	
	// Extract all items with the "result" name and data type.
	bool printed = false;
	int32 type = 0;
	const void* voidptr;
	int32 sizefound;
	while (fReply.FindData(SF_RESULT, typefound, type++, &voidptr, &sizefound) == B_OK)
	{
		char* text = format_data(typefound, (char*)voidptr, sizefound);
		printf("%s\n", text);
		if (strlen(text))
			printed = true;
		delete[] text;
	}
	
	// Make sure we display something meaningful in any case.
	if (!printed)
		puts("Invisible reply!");
}

// Print meaningful error message.
void Command::PrintError(status_t result)
{
	// The strategy is: we begin to looking for a "message" string
	// in the command reply. If that fails, we display the ugly
	// "Unknown Error" message, but at least we give a decent
	// hex number instead of a useless decimal number.
	string strErr(strerror(result));					
	string strCmd(Text());
	BMessage reply(Reply());
	const char* message;
	if (reply.FindString(SF_MESSAGE, &message) == B_OK)
		strErr = message;
	else
	{
		// We don't want to display horrible "Unknown Error" messages
		// in *decimal*, which is really an insult for a programmer...
		// If we're lucky, though, the system error is meaningful enough.
		if (strErr.find("Unknown Error") == 0)
		{
			// Life sucks. Build a generic yet decent error message.
			char buf[64];
			sprintf(buf, "Unknown Error 0x%08lX", result);
			strErr = buf;
		}
	}
	fprintf(stderr, "%s -> %s\n", strCmd.c_str(), strErr.c_str());
}

// Make sure the command is actually usable.
status_t Command::InitCheck() const
{
	if (!fInited)
		return B_NO_INIT;
	if (!fMessenger)
		return B_BAD_HANDLER;
	if (!fMessenger->IsValid())
		return B_BAD_PORT_ID;
	return B_OK;
}

// Return default string representation for this command.
string Command::Text() const
{
	char buf[8];
	return string(id2str(fMessage.what, buf));
}

// Add contents of specifier message (any command).
void Command::AddSpecifier(const BMessage* specifier)
{
	// We cannot just add the whole message as a specifier because apparently
	// the scripting system expects all specifiers to be at the same level
	// in a message -- i.e. no "nesting" of specifiers is allowed.
	int32 item = 0;
	BMessage spec;
	while (specifier->FindMessage(SF_SPECIFIERS, item++, &spec) == B_OK)
		fMessage.AddSpecifier(&spec);
	
	// We're not done yet. If we have an autoresolve request, we have
	// to add it to the message.
	int32 autoresolve;
	status_t error = specifier->FindInt32("autoresolve", &autoresolve);
	if (error == B_OK)
		fMessage.AddInt32("autoresolve", autoresolve);
}

// Adjust the first specifier of a message for special commands.
void Command::AdjustSpecifier(BMessage* specifier, const char* defaultProperty, bool addIt, const char* addProperty)
{
	// Make sure we have a valid kind of specifier for this operation.
	BMessage spec;
	status_t error = specifier->FindMessage(SF_SPECIFIERS, 0, &spec);
	if (error != B_OK)
		return;	
	else if (spec.what == B_DIRECT_SPECIFIER)
	{
		// Scan through the message and reconstruct it in the proper way.
		BMessage newSpecifier(specifier->what);
		if (addIt)
			newSpecifier.AddSpecifier(addProperty);
		int32 item = 0;
		while (specifier->FindMessage(SF_SPECIFIERS, item++, &spec) == B_OK)
		{
			// If we are dealing with the first specifier, we have to translate it,
			// but only if it's a proper specifier.
			if (item == 1)
			{
				const char* name;
				if (spec.FindString(SF_PROPERTY, &name) == B_OK)
				{
					// Turn a direct into a name specifier.
					string direct(name);
					spec.MakeEmpty();
					spec.what = B_NAME_SPECIFIER;
					spec.AddString(SF_PROPERTY, defaultProperty);
					spec.AddString(SF_NAME, direct.c_str());
				}
			}
			newSpecifier.AddSpecifier(&spec);
		}
		
		// We're not done yet. If we have an autoresolve request, we have
		// to add it to the message.
		int32 autoresolve;
		status_t error = specifier->FindInt32("autoresolve", &autoresolve);
		if (error == B_OK)
		{
			// Make sure to push the autoresolve item down if we have to add
			// a direct specifier.
			if (addIt)
				++autoresolve;
			newSpecifier.AddInt32("autoresolve", autoresolve);
		}

		// Adjust specifier.
		*specifier = newSpecifier;
	}
}

// Add contents of data message (mostly the SET command, maybe others).
void Command::AddData(const BMessage* data)
{
	// Dehydrate data message in command message.
	// We do not recur over messages or flattened objects -- this method is meant
	// to add "pure" data, under any type and name, to the command message.
	// As for the above method, just adding the message instead of its contents
	// would not work with the scripting system.
	int32 count = data->CountNames(B_ANY_TYPE);
	for (int32 i = 0; i < count; ++i)
	{
		// Identify name and data type.
		ulong typefound;
		char* namefound;
		data->GetInfo(B_ANY_TYPE, i, &namefound, &typefound);
		
		// Extract all items with the current name and data type.
		int32 type = 0;
		const void *voidptr;
		int32 sizefound;
		while (data->FindData(namefound, typefound, type++, &voidptr, &sizefound) == B_OK)
			fMessage.AddData(namefound, typefound, voidptr, sizefound);
	}
}

// Actual command execution.
status_t Command::Execute()
{
	// Make sure we can run this command.
	if (InitCheck() != B_OK)
		return B_BAD_HANDLER;
	
	// Run it: if execution fails, give up immediately.
	status_t error = fMessenger->SendMessage(&fMessage, &fReply);
	if (error != B_OK)
		return error;
	
	// Interpret reply properly: report an "error" field if present.
	fSent = true;
	if (fReply.what == B_REPLY)
	{
		int32 errorCode = B_NO_ERROR;
		error = fReply.FindInt32(SF_ERROR, &errorCode);
		return (error == B_OK) ? errorCode : B_OK;
	}
	else if (fReply.what == B_NO_REPLY)
		return B_OK;
	else if (fReply.what != B_NO_ERROR)
		return fReply.what;
	else
		return error;
}

// Attempt automatic resolution of specifier if possible.
status_t Command::AutoResolve(status_t error)
{
	if (*gVerbosity > 2)
		printf("AutoResolve: %s\n", Text().c_str());
	// Make sure we have an autoresolve item, otherwise the attempt is pointless.
	// This ensures that autoresolution is requested from the second item only.
	int32 autoresolve;
	status_t result = fMessage.FindInt32("autoresolve", &autoresolve);
	if (result != B_OK || autoresolve < 1)
	{
		// We're apparently out of luck: no dynamic specifier was given explicitly.
		// We now check if the command has a parent: in this case, we can attempt
		// a blind resolution.
		if (!fParent)
			return error;
		fDynamic = fParent->Dynamic();
		if (!fDynamic)
		{
			result = fParent->BlindSolve();
			if (result != B_OK)
				return error;
			fDynamic = fParent->Dynamic();
			int32 numViews = fDynamic->Analyze();
			if (numViews < 1)
				return error;
		}
		
		// Extract information needed to attempt the search.
		int32 curSpec = -1;
		BMessage specMsg(B_NO_SPECIFIER);
		result = fMessage.GetCurrentSpecifier(&curSpec, &specMsg);
		if (result != B_OK)
			return error;
		if (specMsg.what == B_NO_SPECIFIER || curSpec < 0)
			return error;
		const char* property;
		result = specMsg.FindString(SF_PROPERTY, &property);
		if (result != B_OK)
			return error;
		const char* name;
		result = specMsg.FindString(SF_NAME, &name);
		if (result != B_OK)
			return error;
		
		// Attempt search.
		ScriptingElement* element = fDynamic->FindElement(name);
		if (!element)
			return error;
		
		// We got it!
		// Now we look for the specifier immediately before the target item.
		// That is at a distance of one specifier, if it is present at all.
		BMessage original(B_NO_SPECIFIER);
		if (curSpec >= 1)
			fMessage.FindMessage(SF_SPECIFIERS, curSpec - 1, &original);
		
		// Finally, purge current specifier and reconstruct it.
		// Possibly add data fields too. Then, re-execute the command.
		// Note that data must be added with what != B_MESSAGE_TYPE.
		
		fMessage.MakeEmpty();
		if (original.what != B_NO_SPECIFIER)
			fMessage.AddSpecifier(&original);
		if (element->GetAccessPtr())
			AddSpecifier(element->GetAccessPtr());
		if (fValue.what != B_MESSAGE_TYPE)
			AddData(&fValue);
			
		return Execute();
	}
		
	// Locate the autoresolving item and the target specifier. This is located
	// in the specifier immediately preceding the autoresolve item.
	BMessage autoMsg;
	result = fMessage.FindMessage(SF_SPECIFIERS, autoresolve, &autoMsg);
	if (result != B_OK)
		return result;
	BMessage targetMsg;
	result = fMessage.FindMessage(SF_SPECIFIERS, autoresolve - 1, &targetMsg);
	if (result != B_OK)
		return result;
	const char* targetName;
	result = targetMsg.FindString(SF_NAME, &targetName);
	if (result != B_OK)
		return result;
	
	// Make sure it is a valid specifier.
	// Currently we accept only "in Window {name,index}".
	// Hopefully we'll also be able to resolve views in the future.
	if (autoMsg.what != B_NAME_SPECIFIER && autoMsg.what != B_INDEX_SPECIFIER)
		return error;
	const char* property;
	if (autoMsg.FindString(SF_PROPERTY, &property) != B_OK)
		return error;
	if (strcmp(property, SM_WINDOW))
		return error;
	
	// Extract the specifier type and build scripting window.
	// Only two types of window can be created at the moment.
	ScriptingWindow* window = 0;
	int32 index;
	const char* name;
	BMessenger* msgr = fMessenger->Messenger();
	if (autoMsg.what == B_NAME_SPECIFIER)
	{
		result = autoMsg.FindString(SF_NAME, &name);
		window = new ScriptingWindow(name, new BMessenger(*msgr));
	}
	else
	{
		result = autoMsg.FindInt32(SF_INDEX, &index);
		window = new ScriptingWindow(index, new BMessenger(*msgr));
	}
	
	// Turn paranoia mode on.
	if (result != B_OK)
	{
		delete window;
		return result;
	}
	
	// Delegate specifier solution.
	return Solve(window, targetName, autoresolve, error);
}

// Solve a dynamic specifier.
status_t Command::Solve(	ScriptingElement* target,
							const char* targetName,
							int32 autoresolve,
							status_t error)
{
	// Analyze target and resolve specifier. Currently this is a little slow
	// because it performs a complete analysis. The next version will stop
	// at the object being resolved.
	int32 numViews = target->Analyze();
	if (numViews > 0)
	{
		ScriptingElement* element = target->FindElement(targetName);
		if (element)
		{
			// We got it!
			// Now we look for the specifier immediately before the target item.
			// That is at a distance of two specifiers, if it is present at all.
			// This implies that currently only specifiers such as "View x in Window y"
			// will be understood properly.
			BMessage original(B_NO_SPECIFIER);
			if (autoresolve >= 2)
				fMessage.FindMessage(SF_SPECIFIERS, autoresolve - 2, &original);
			
			// Finally, purge current specifier and reconstruct it.
			// Possibly add data fields too. Then, re-execute the command.
			// Note that data must be added with what != B_MESSAGE_TYPE.
			
			fMessage.MakeEmpty();
			if (original.what != B_NO_SPECIFIER)
				fMessage.AddSpecifier(&original);
			if (element->GetAccessPtr())
				AddSpecifier(element->GetAccessPtr());
			if (fValue.what != B_MESSAGE_TYPE)
				AddData(&fValue);

			delete target;
			return Execute();
		}
	}
	
	// If we fall through here, we have to give up. The view cannot be resolved at all,
	// not even dynamically. This might also mean that the specifier was invalid.
	delete target;
	return error;
}

// Blindly solve a specifier.
status_t Command::BlindSolve()
{
	// Analyze message: extract current specifier and its property.
	int32 curSpec = -1;
	BMessage specMsg(B_NO_SPECIFIER);
	status_t error = fMessage.GetCurrentSpecifier(&curSpec, &specMsg);
	if (error != B_OK)
		return error;
	if (specMsg.what == B_NO_SPECIFIER || curSpec < 0)
		return B_NO_SPECIFIER;
	const char* property;
	error = specMsg.FindString(SF_PROPERTY, &property);
	if (error != B_OK)
		return error;
	
	// Make sure we have a valid object and create scripting target.
	if (strcmp(property, SM_WINDOW))
		return B_NAME_NOT_FOUND;
	
	// Interpret specifier.
	BMessenger* msgr = Messenger()->Messenger();
	status_t result = B_OK;
	const char* name;
	int32 index;
	if (specMsg.what == B_NAME_SPECIFIER)
	{
		result = specMsg.FindString(SF_NAME, &name);
		fDynamic = new ScriptingWindow(name, new BMessenger(*msgr), 0, 0, false);
	}
	else
	{
		result = specMsg.FindInt32(SF_INDEX, &index);
		fDynamic = new ScriptingWindow(index, new BMessenger(*msgr), 0, 0, false);
	}
	
	// Turn paranoia mode on.
	if (result != B_OK)
	{
		delete fDynamic;
		fDynamic = 0;
		return result;
	}
	return B_OK;
}

// A list of commands represents a block of commands for the same target.
CommandList::CommandList() : list<Command*>() {}

// Yes, we could use the standard algorithms to iterate over the list...
// but we'll do that when we know that the list contents are safe to handle.
// Right now, we'll resort to RTTI and tighter control.
status_t CommandList::Do()
{
	CommandList::iterator curCommand = begin();
	for (; curCommand != end(); ++curCommand)
	{
		Command* command = dynamic_cast<Command*>(*curCommand);
		if (command)
		{
			// Check verbosity levels and execute the current command.
			if (*gVerbosity > 0)
				printf("Execute: %s\n", command->Text().c_str());
			if (*gVerbosity > 1)
				command->PrintContents();
			status_t result = command->Do();
			if (result != B_OK)
			{
				// We display errors in one copy only, e.g., not for a TELL command,
				// otherwise we'd get duplicate errors.
				TellCommand* tellCmd = dynamic_cast<TellCommand*>(command);
				if (!tellCmd)
					command->PrintError(result);
				return result;
			}
			else
			{
				// Currently we waste a few millisecs per command execution
				// in order not to overload the CPU(s).
				snooze(1000);
			}
			if (*gVerbosity > 2)
			{
				printf("Result: %s => ", command->Text().c_str());
				command->PrintResult();
			}
		}
	}
	return B_OK;
}

void CommandList::Print(int indent = 0)
{
	for (CommandList::iterator cmd = begin(); cmd != end(); ++cmd)
	{
		Command* command = dynamic_cast<Command*>(*cmd);
		if (command)
			command->Print(indent);
	}
}

void CommandList::PrintContents(bool printReply = true)
{
	for (CommandList::iterator cmd = begin(); cmd != end(); ++cmd)
	{
		Command* command = dynamic_cast<Command*>(*cmd);
		if (command)
			command->PrintContents(printReply);
	}
}

// Actual commands.

QuitCommand::QuitCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_QUIT_REQUESTED)
{
	AddSpecifier(specifier);
}

string QuitCommand::Text() const
{
	string str("quit");
	if (fMessage.HasSpecifiers())
	{
		str += " ";
		str += specifier_to_str(&fMessage);
	}
	return str;
}

AboutCommand::AboutCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_ABOUT_REQUESTED)
{
	AddSpecifier(specifier);
}

string AboutCommand::Text() const
{
	string str("about");
	if (fMessage.HasSpecifiers())
	{
		str += " ";
		str += specifier_to_str(&fMessage);
	}
	return str;
}

SaveCommand::SaveCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_SAVE_REQUESTED)
{
	AddSpecifier(specifier);
}

string SaveCommand::Text() const
{
	string str("save");
	if (fMessage.HasSpecifiers())
	{
		str += " ";
		str += specifier_to_str(&fMessage);
	}
	if (!fValue.IsEmpty())
	{
		str += " with ";
		str += data_to_str(&fValue);
	}
	return str;
}

LoadCommand::LoadCommand(DMessenger* msgr, const char* file)
: Command(msgr, B_REFS_RECEIVED)
{
	status = AddFile(file);
}

LoadCommand::LoadCommand(DMessenger* msgr, const entry_ref* file)
: Command(msgr, B_REFS_RECEIVED)
{
	status = AddFile(file);
}

status_t LoadCommand::InitCheck() const
{
	status_t error = Command::InitCheck();
	return error == B_OK ? status : error;
}

string LoadCommand::Text() const
{
	string str("load file(\"");
	str += path;
	return str + "\")";
}

status_t LoadCommand::AddFile(const char* file)
{
	path = file;
	BEntry entry(file);
	entry_ref ref;
	status_t error = entry.GetRef(&ref);
	if (error == B_OK)
		return AddFile(&ref);
	else
	{
		path += " (invalid!)";
		return error;
	}
}

status_t LoadCommand::AddFile(const entry_ref* file)
{
	// Compute full file path.
	BPath fullPath(file);
	path = fullPath.Path();

	// We have to add the file twice, because B_REFS_RECEIVED expects it
	// under "refs" whereas scripting expects it under "data".
	status_t error = fMessage.AddRef(SF_REFS, file);
	return (error == B_OK) ? fMessage.AddRef(SF_DATA, file) : error;
}

GetSuitesCommand::GetSuitesCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_GET_SUPPORTED_SUITES) 
{
	AddSpecifier(specifier);
}

status_t GetSuitesCommand::Do()
{
	status_t result = Command::Do();
	if (result != B_OK)
		return result;

	//	Iterate over suites and stuff them in a list.
	BMessage suites(fReply);
	BList list;
	int32 current = 0;
	while (true)
	{
		//	Locate the suite name first.		
		const char* suiteName;
		status_t error = suites.FindString(SF_SUITES, current, &suiteName);
		if (error != B_OK)
			break;
		
		//	Locate the suite properties and add them to the list.		
		printf("Supports %s:\n", suiteName);
		int32 sizefound;
		const void* voidptr;
		error = suites.FindData(SF_MESSAGES, B_PROPERTY_INFO_TYPE, current, &voidptr, &sizefound);
		if (error != B_OK)
			break;

		BPropertyInfo propinfo;
		if (propinfo.Unflatten(B_PROPERTY_INFO_TYPE, voidptr, sizefound) == B_OK)
			FullDehydratePropertyInfo(&list, propinfo, sizefound, false);
		
		//	Now we display a suitable list.
		int32 lines = list.CountItems();
		for (int32 line = 0; line < lines; ++line)
		{
			char* lineText = (char*)list.ItemAt(line);
			printf("%s\n", lineText);
			delete[] lineText;
		}
		list.MakeEmpty();

		//	Go over all available suites.		
		++current;
	}
	return B_OK;
}

string GetSuitesCommand::Text() const
{
	string str("getsuites");
	if (fMessage.HasSpecifiers())
	{
		str += " ";
		str += specifier_to_str(&fMessage);
	}
	return str;
}

GetCommand::GetCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_GET_PROPERTY) 
{
	AddSpecifier(specifier);
}

status_t GetCommand::Do()
{
	return DoAndPrintReply();
}

string GetCommand::Text() const
{
	string str("get ");
	str += specifier_to_str(&fMessage);
	return str;
}

SetCommand::SetCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value)
: Command(msgr, B_SET_PROPERTY)
{
	fValue = *value;
	AddSpecifier(specifier);
	AddData(&fValue);
}

string SetCommand::Text() const
{
	string str("set ");
	str += specifier_to_str(&fMessage);
	str += " to ";
	str += data_to_str(&fValue);
	return str;
}

DoCommand::DoCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_EXECUTE_PROPERTY) 
{
	AddSpecifier(specifier);
}

DoCommand::DoCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value)
: Command(msgr, B_EXECUTE_PROPERTY) 
{
	fValue = *value;
	AddSpecifier(specifier);
	AddData(&fValue);
}

// Custom commands might have to be extracted and analyzed.
status_t DoCommand::Do()
{
	// First of all, attempt normal execution.
	status_t error = Command::Do();
	if (error == B_OK)
		return error;
	
	// This might be a custom command. Let's look for it.
	// If the first specifier contains a suitable property, we can try to search it.
	BMessage spec;
	error = fMessage.FindMessage(SF_SPECIFIERS, 0, &spec);
	if (error != B_OK)
		return error;
	
	// Determine property and collect other specifier information.
	// Currently we accept also name and index specifiers besides the classic
	// direct specifier that is most often used in this context. However,
	// we are not using the other two types of specifiers yet.
	const char* property;
	error = spec.FindString(SF_PROPERTY, &property);
	if (error != B_OK)
		return error;
	const char* name;
	int32 index;
	switch (spec.what)
	{
		case B_DIRECT_SPECIFIER:
		break;
		case B_NAME_SPECIFIER:
			error = spec.FindString(SF_NAME, &name);
			if (error != B_OK)
				return error;
		break;
		case B_INDEX_SPECIFIER:
			error = spec.FindInt32(SF_INDEX, &index);
			if (error != B_OK)
				return error;
		break;
		default:
			return B_BAD_SCRIPT_SYNTAX;
	}

	// Obtain suites of object and locate custom command.
	BMessage reply;
	BMessage suites(B_GET_SUPPORTED_SUITES);
	bool found = false;
	uint32 what = 0;
	if (fMessenger->SendMessage(&suites, &reply) == B_OK)
	{
		// Locate all properties supported by this object.
		int32 j = 0;
		const void* voidptr;
		int32 sizefound;
		BPropertyInfo propinfo;
		const value_info* vinfo;
		int32 vinfo_index;
		int32 vinfo_count;
		
		while (reply.FindData(	SF_MESSAGES,
								B_PROPERTY_INFO_TYPE,
								j++,
								&voidptr,
								&sizefound) == B_OK
				&& !found)
		{
			// Retrieve a property from the object.
			if (propinfo.Unflatten(B_PROPERTY_INFO_TYPE, voidptr, sizefound) == B_OK)
			{
				vinfo = propinfo.Values();
				vinfo_index = 0;
				vinfo_count = propinfo.CountValues();

				// Check whether we know about this property.
				while (vinfo_index < vinfo_count)
				{
					if (strcmp(vinfo[vinfo_index].name, property) == 0)
					{
						found = true;
						what = vinfo[vinfo_index].value;
						break;
					}
					++vinfo_index;
				}
			}
		}
	}
	
	// Final check.
	if (!found)
		return B_BAD_SCRIPT_SYNTAX;
	
	// We have found the custom verb. We have to strip the first specifier
	// from the original message and then reconstruct it properly.
	
	BMessage oldSpec(fMessage);
	fMessage.MakeEmpty();
	fMessage.what = what;
	int32 item = 1;
	while (oldSpec.FindMessage(SF_SPECIFIERS, item++, &spec) == B_OK)
		fMessage.AddSpecifier(&spec);
	
	// We're not done yet. If we have an autoresolve request, we have
	// to add it to the message. We also have to add data (if any).
	int32 autoresolve;
	error = oldSpec.FindInt32("autoresolve", &autoresolve);
	if (error == B_OK)
		fMessage.AddInt32("autoresolve", autoresolve);
	if (!fValue.IsEmpty())
		AddData(&fValue);
	
	// We can now attempt a re-execution of the updated command.
	return Command::Do();
}

string DoCommand::Text() const
{
	string str("do ");
	str += specifier_to_str(&fMessage);
	if (!fValue.IsEmpty())
	{
		str += " ";
		str += data_to_str(&fValue);
	}
	return str;
}

DeleteCommand::DeleteCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_DELETE_PROPERTY) 
{
	AddSpecifier(specifier);
}

string DeleteCommand::Text() const
{
	string str("delete ");
	str += specifier_to_str(&fMessage);
	return str;
}

CountCommand::CountCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_COUNT_PROPERTIES) 
{
	AddSpecifier(specifier);
}

status_t CountCommand::Do()
{
	return DoAndPrintReply();
}

string CountCommand::Text() const
{
	string str("count ");
	str += specifier_to_str(&fMessage);
	return str;
}

CreateCommand::CreateCommand(DMessenger* msgr, const BMessage* specifier)
: Command(msgr, B_CREATE_PROPERTY) 
{
	AddSpecifier(specifier);
}

CreateCommand::CreateCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value)
: Command(msgr, B_CREATE_PROPERTY)
{
	fValue = *value;
	AddSpecifier(specifier);
	AddData(&fValue);
}

string CreateCommand::Text() const
{
	string str("create ");
	str += specifier_to_str(&fMessage);
	if (!fValue.IsEmpty())
	{
		str += " with ";
		str += data_to_str(&fValue);
	}
	return str;
}

// Retains ownership of passed messenger and command list.
// Note that the actual message for this command will never be executed.
TellCommand::TellCommand(DMessenger* msgr, string& target, CommandList* block)
: Command(msgr, B_EXECUTE_PROPERTY), fTarget(target), fCommands(block) {}

TellCommand::~TellCommand()
{
	if (fCommands)
		fCommands->clear();
	delete fCommands;
	delete fMessenger;
}

status_t TellCommand::Do()
{
	// We cannot just run the command list blindly -- i.e. if we have an autoresolving
	// specifier, we might try to access a bad messenger, and this would result
	// in a crash. We must InitCheck() the object and only then possibly pass the
	// execution message to the list.
	status_t error = InitCheck();
	if (error != B_OK)
	{
		PrintError(error);
		return error;
	}
	return fCommands->Do();
}

void TellCommand::Print(int indent = 0)
{
	PrintIndentation(indent);
	printf("%s\n", Text().c_str());
	fCommands->Print(indent + 1);
	PrintIndentation(indent);
	puts("endtell");
}

void TellCommand::PrintContents(bool printReply = true)
{
	fCommands->PrintContents(printReply);
}

string TellCommand::Text() const
{
	return string("tell ") + fTarget;
}

// Retains ownership of passed command list.
LetCommand::LetCommand(DMessenger* msgr, BMessage* specifier, CommandList* block)
: Command(msgr, B_GET_PROPERTY), fCommands(block)
{
	// Make sure we update the autoresolve field if present. This is needed because
	// we're asking for the messenger of the given specifier, and later we'll be
	// using it for all commands.
	int32 autoresolve;
	status_t error = specifier->FindInt32("autoresolve", &autoresolve);
	if (error == B_OK)
		specifier->ReplaceInt32("autoresolve", autoresolve + 1);

	// Add specifier and update messenger for target command.
	// We must be flexible here: we add a messenger specifier only if needed.
	// This is to balance the current inability to express partial specifiers.	
	bool addIt = true;
	BMessage spec;
	error = specifier->FindMessage(SF_SPECIFIERS, 0, &spec);
	if (error == B_OK)
	{
		const char* property;
		error = spec.FindString(SF_PROPERTY, &property);
		if (error == B_OK)
		{
			switch (spec.what)
			{
				case B_DIRECT_SPECIFIER:
				{
					// A direct specifier. This is a problem! The choice is:
					// either autocompletion or custom specifiers...
					// Currently we support the latter, so we cancel the addition
					// of a messenger specifier.
					addIt = false;
				}
				break;
				case B_NAME_SPECIFIER:
				{
					// If the property isn't known to us, we cancel the messenger.
					// We do this because most likely this specifier will return
					// a messenger by itself -- the only two ones we know about
					// are Window and View, which both need an explicit messenger.
					if (strcmp(property, SM_WINDOW) && strcmp(property, SM_VIEW))
						addIt = false;
				}
				break;
			}
		}
	}
	
	// Prepare specifier to fire.
	if (addIt)
		fMessage.AddSpecifier(SM_MESSENGER);
	AddSpecifier(specifier);
}

LetCommand::~LetCommand()
{
	// If there are commands available, we have a new messenger that can be deleted.
	// Otherwise, we cannot delete it because it probably belongs to the parent command.
	if (fCommands)
	{
		fCommands->clear();
		delete fMessenger;
	}
	delete fCommands;
	
	// Delete the dynamic specifier that might have been created automatically.
	delete fDynamic;
}

status_t LetCommand::Do()
{
	// First we resolve the specifier for the target commands, by asking
	// for the messenger of the target.
	// This of course will be done dynamically if possible and needed.
	status_t error = Command::Do();
	if (error != B_OK)
	{
		// Ensure we setup an error condition if something goes wrong.
		delete fCommands;
		fCommands = 0;
		return error;
	}
	
	// Extract the specifier and make it the new target messenger.
	BMessenger msgr;
	error = fReply.FindMessenger(SF_RESULT, &msgr);
	if (error != B_OK)
	{
		// Ensure we setup an error condition if something goes wrong.
		delete fCommands;
		fCommands = 0;
		return error;
	}
	else
	{
		// It is a go! Create a new messenger for this command.
		// Then change messenger and parent for all commands in the list.
		// Once again, we're in paranoia mode therefore we dynamic_cast.
		DMessenger* messenger = new DMessenger(&msgr);
		CommandList::iterator cmd = fCommands->begin();
		for (; cmd != fCommands->end(); ++cmd)
		{
			Command* command = dynamic_cast<Command*>(*cmd);
			if (command)
			{
				command->SetMessenger(messenger);
				command->SetParent(this);
			}
		}
		
		// Adjust messenger and run the command list.
		SetMessenger(messenger);
		return fCommands->Do();
	}
}

void LetCommand::Print(int indent = 0)
{
	PrintIndentation(indent);
	printf("%s\n", Text().c_str());
	if (fCommands)
		fCommands->Print(indent + 1);
	if (fCommands && fCommands->size() > 1)
	{
		PrintIndentation(indent);
		puts("end");
	}
}

void LetCommand::PrintContents(bool printReply = true)
{
	if (fCommands)
		fCommands->PrintContents(printReply);
}

status_t LetCommand::InitCheck()
{
	return fCommands ? Command::InitCheck() : B_ERROR;
}

string LetCommand::Text() const
{
	string str("let ");
	if (fCommands && fCommands->size() < 2)
		str = "tell ";
	str += specifier_to_str(&fMessage);
	if (fCommands && fCommands->size() > 1)
		str += " do";
	else
		str += " to";
	return str;
}

LaunchCommand::LaunchCommand(DMessenger* msgr, const char* application)
: Command(msgr, B_SOME_APP_LAUNCHED), fApplication(application) {}

status_t LaunchCommand::Do()
{
	// Locate application.
	entry_ref ref;
	status_t error = be_roster->FindApp(fApplication.c_str(), &ref);
	if (error == B_OK)
		error = be_roster->Launch(&ref);
	
	// We also accept as valid an already running application, but we add it
	// to the message only if it's a fresh launch. In fact, the message
	// won't be sent if there is no valid launch.
	if (error == B_OK)
		error = fMessage.AddRef(SF_REFS, &ref);
	else
		fMessage.what = 0;
	if (error != B_OK && error != B_ALREADY_RUNNING)
	{
		// Destroy the messenger to signal an error condition.
		// This will cause InitCheck() to fail.
		delete fMessenger;
		fMessenger = 0;
		return error;
	}
	else
	{
		// Wait for a while -- until the application has been launched --
		// but not indefinitely, at most five seconds. Margin of error
		// in waiting time is around one millisecond.
		bigtime_t systime = system_time();
		while (be_roster->IsRunning(fApplication.c_str()) == false)
		{
			snooze(1000);
			if (system_time() - systime > 5000000)
				break;
		}
		bool running = be_roster->IsRunning(fApplication.c_str());
		return running ? B_OK : B_LAUNCH_FAILED;
	}
}

string LaunchCommand::Text() const
{
	string str("launch \"");
	str += fApplication;
	return str + "\"";
}

TypeCommand::TypeCommand(DMessenger* msgr, BMessage* specifier, const BMessage* value)
: Command(msgr, B_SET_PROPERTY)
{
	fValue = *value;
	AdjustSpecifier(specifier);
	AddSpecifier(specifier);
	AddData(&fValue);
}

string TypeCommand::Text() const
{
	string str("type ");
	str += data_to_str(&fValue);
	str += " into ";
	str += specifier_to_str(&fMessage, true);
	return str;
}

ClickCommand::ClickCommand(DMessenger* msgr, BMessage* specifier)
: Command(msgr, B_SET_PROPERTY)
{
	fValue = BMessage(B_INT32_TYPE);
	fValue.AddInt32(SF_DATA, B_CONTROL_ON);
	AdjustSpecifier(specifier);
	AddSpecifier(specifier);
	AddData(&fValue);
}

string ClickCommand::Text() const
{
	string str("click ");
	str += specifier_to_str(&fMessage, true);
	return str;
}

VerbCommand::VerbCommand(DMessenger* msgr, uint32 verb, BMessage* specifier)
: Command(msgr, verb)
{
	AddSpecifier(specifier);
}

VerbCommand::VerbCommand(DMessenger* msgr, uint32 verb, BMessage* specifier, BMessage* value)
: Command(msgr, verb)
{
	fValue = *value;
	AddSpecifier(specifier);
	AddData(&fValue);
}

string VerbCommand::Text() const
{
	string str("do ");
	char verb[8];
	str += id2str(fMessage.what, verb);
	str += specifier_to_str(&fMessage);
	if (!fValue.IsEmpty())
	{
		str += " with ";
		str += data_to_str(&fValue);
	}
	return str;
}

// CHOOSE does not accept dynamic specifiers because it is targeted at menus only.
ChooseCommand::ChooseCommand(DMessenger* msgr, BMessage* specifier)
: Command(msgr, B_EXECUTE_PROPERTY)
{
	// Inspect the current specifier.
	int32 curSpec = -1;
	status_t error = specifier->GetCurrentSpecifier(&curSpec);
	if (error != B_OK || curSpec < 0)
	{
		SetInited(false);
		return;
	}
	
	// Now explore the specifier: the first one is supposed to be direct,
	// and the last one might be a window. The MenuBar must *not* be present,
	// because it is added here. MenuItem must not be present either.
	bool added = false;
	int32 index = 0;
	BMessage specMsg(B_NO_SPECIFIER);
	while ((error = specifier->FindMessage(SF_SPECIFIERS, index, &specMsg)) == B_OK)
	{
		//	Locate property.
		const char* property;
		error = specMsg.FindString(SF_PROPERTY, &property);
		if (error != B_OK)
		{
			SetInited(false);
			return;
		}
		if (index == 0)
		{
			// The first one must be direct.
			if (specMsg.what != B_DIRECT_SPECIFIER)
			{
				SetInited(false);
				return;
			}
			
			// Add adjusted specifier.
			fMessage.AddSpecifier(SM_MENUITEM, property);
		}
		else if (index == curSpec)
		{
			// The last specifier might not be a Menu: in this case, we first add
			// a MenuBar specifier, then we add the object.
			if (strcmp(property, SM_MENU))
			{
				added = true;
				fMessage.AddSpecifier(SM_MENUBAR);
				fMessage.AddSpecifier(&specMsg);
			}
			else
			{
				// It is a regular specifier (e.g. a Menu "SomeMenu"). Add it
				// first, then add the MenuBar.
				added = true;
				fMessage.AddSpecifier(&specMsg);
				fMessage.AddSpecifier(SM_MENUBAR);
			}
		}
		else
		{
			// A regular specifier (e.g. Menu "SomeMenu") is added directly.
			fMessage.AddSpecifier(&specMsg);
		}
		
		// Examine all available specifiers.
		++index;
	}
	
	// Make sure we have added the MenuBar object. This implies that CHOOSE
	// cannot be issued within a LET block addressed to a MenuBar, but the
	// reason the command exists is exactly to avoid such blocks.
	if (!added)
		fMessage.AddSpecifier(SM_MENUBAR);
}

// This constructor is meant for accessing items by index.
ChooseCommand::ChooseCommand(DMessenger* msgr, int32 item, BMessage* specifier)
: Command(msgr, B_EXECUTE_PROPERTY)
{
	// Inspect the current specifier.
	int32 curSpec = -1;
	status_t error = specifier->GetCurrentSpecifier(&curSpec);
	if (error != B_OK || curSpec < 0)
	{
		SetInited(false);
		return;
	}
	
	// Now explore the specifier: The MenuBar must *not* be present,
	// because it is added here. MenuItem must not be present either.
	bool added = false;
	int32 index = 0;
	BMessage specMsg(B_NO_SPECIFIER);
	while ((error = specifier->FindMessage(SF_SPECIFIERS, index, &specMsg)) == B_OK)
	{
		//	Locate property.
		const char* property;
		error = specMsg.FindString(SF_PROPERTY, &property);
		if (error != B_OK)
		{
			SetInited(false);
			return;
		}
		if (index == 0)
		{
			// Add item specifier plus current specifier.
			fMessage.AddSpecifier(SM_MENUITEM, item);
			fMessage.AddSpecifier(&specMsg);
		}
		else if (index == curSpec)
		{
			// The last specifier might not be a Menu: in this case, we first add
			// a MenuBar specifier, then we add the object.
			if (strcmp(property, SM_MENU))
			{
				added = true;
				fMessage.AddSpecifier(SM_MENUBAR);
				fMessage.AddSpecifier(&specMsg);
			}
			else
			{
				// It is a regular specifier (e.g. a Menu "SomeMenu"). Add it
				// first, then add the MenuBar.
				added = true;
				fMessage.AddSpecifier(&specMsg);
				fMessage.AddSpecifier(SM_MENUBAR);
			}
		}
		else
		{
			// A regular specifier (e.g. Menu "SomeMenu") is added directly.
			fMessage.AddSpecifier(&specMsg);
		}
		
		// Examine all available specifiers.
		++index;
	}
	
	// Make sure we have added the MenuBar object. This implies that CHOOSE
	// cannot be issued within a LET block addressed to a MenuBar, but the
	// reason the command exists is exactly to avoid such blocks.
	if (!added)
		fMessage.AddSpecifier(SM_MENUBAR);
}

string ChooseCommand::Text() const
{
	string str("choose ");
	str += specifier_to_str(&fMessage);
	return str;
}

PauseCommand::PauseCommand(DMessenger* msgr, int32 pause)
: Command(msgr, 0), fPause(bigtime_t(pause)) {}

status_t PauseCommand::Do()
{
	bigtime_t delay = fPause * 1000;
	snooze(delay);
	return B_OK;
}

string PauseCommand::Text() const
{
	char buf[64];
	sprintf(buf, "%ldms", (long)(fPause));
	string str("pause ");
	str += buf;
	return str;
}

PrintCommand::PrintCommand(DMessenger* msgr, const char* str)
: Command(msgr, 0), fString(str) {}

status_t PrintCommand::Do()
{
	printf("%s", fString.c_str());
	return B_OK;
}

string PrintCommand::Text() const
{
	string str("print \"");
	
	// We have to escape the string.
	string s;
	int size = int(fString.size());
	for (int index = 0; index < size; ++index)
	{
		char cur = fString[index];
		switch (cur)
		{
			case '\n': s += "\\n"; break;
			case '\r': s += "\\r"; break;
			case '\t': s += "\\t"; break;
			default: s += cur; break;
		}
	}
	
	// Complete string to display.
	str += s;
	str += "\"";
	return str;
}

AssignCommand::AssignCommand(DMessenger* msgr, BMessage* specifier, const BMessage* value)
: Command(msgr, B_SET_PROPERTY)
{
	fValue = *value;
	AdjustSpecifier(specifier);
	AddSpecifier(specifier);
	AddData(&fValue);
}

string AssignCommand::Text() const
{
	string str("assign ");
	str += data_to_str(&fValue);
	str += " to ";
	str += specifier_to_str(&fMessage);
	return str;
}
