// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Abstract syntax.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

#ifndef ABSTRACTSYNTAX_H
#define ABSTRACTSYNTAX_H

// System headers.

#ifndef __SGI_STL_LIST_H
#include <list>
#endif
#ifndef __STRING__
#include <string>
#endif
#ifndef _MESSENGER_H
#include <Messenger.h>
#endif
#ifndef _BE_LIST_H
#include <List.h>
#endif
#ifndef _PATH_H
#include <Path.h>
#endif

// Referenced classes.

class ScriptingElement;

// A dynamic messenger, capable of resolving itself if initialized with a string identifier.

class DMessenger
{
public:
	DMessenger(const char* identifier);
	DMessenger(DMessenger* messenger);
	DMessenger(BMessenger* messenger);
	~DMessenger();

	status_t			SendMessage(BMessage* msg, BMessage* reply);
	bool				IsValid();
	
	inline BMessenger*	Messenger()	{ return fMessenger; }
	static BMessenger*	ResolveTarget(const char* str);

private:
	BMessenger* fMessenger;
	string		fIdentifier;
};

// Abstract syntax:
// A generic scripting command.

class Command
{
public:
	Command();
	Command(DMessenger* msgr, const uint32 command);
	virtual ~Command();
	
	// Override these to do anything serious.
	virtual status_t Do();
	virtual status_t DoAndPrintReply();
	virtual void PrintContents(bool printReply = true);
	virtual void Print(int indent = 0);
	virtual void PrintIndentation(int indent = 0);
	virtual void PrintResult();
	virtual void PrintError(status_t result);
	virtual status_t InitCheck() const;
	virtual string Text() const;

	inline Command*			Parent() const		{ return fParent; }
	inline ScriptingElement* Dynamic() const	{ return fDynamic; }
	inline DMessenger* 		Messenger() const	{ return fMessenger; }
	inline const BMessage&	Message() const		{ return fMessage; }
	inline const BMessage&	Reply() const		{ return fReply; }
	inline bool				Sent() const		{ return fSent; }
	inline bool				Inited() const		{ return fInited; }
	
	inline void SetMessenger(DMessenger* msgr)	{ fMessenger = msgr; }
	inline void SetParent(Command* parent)		{ fParent = parent; }
	inline void SetDynamic(ScriptingElement* d)	{ fDynamic = d; }
	inline void SetInited(bool inited)			{ fInited = inited; }

protected:
	DMessenger*			fMessenger;
	BMessage			fMessage;
	BMessage			fReply;
	BMessage			fValue;
	Command*			fParent;
	ScriptingElement*	fDynamic;
	void				AddSpecifier(const BMessage* specifier);
	void				AdjustSpecifier(BMessage* specifier,
										const char* defaultProperty = "View",
										bool addIt = true,
										const char* addProperty = "Value");
	void				AddData(const BMessage* data);

private:
	bool		fSent;
	bool		fInited;
	status_t	Execute();
	status_t	AutoResolve(status_t error);
	status_t	Solve(	ScriptingElement* target,
						const char* targetName,
						int32 autoresolve,
						status_t error);
	status_t	BlindSolve();
};

// A list of commands represents a block of commands for the same target.

class CommandList : public list<Command*>
{
public:
	CommandList();
	status_t Do();
	void Print(int indent = 0);
	void PrintContents(bool printReply = true);
};

// Actual commands.

class QuitCommand : public Command
{
public:
	QuitCommand(DMessenger* msgr) : Command(msgr, B_QUIT_REQUESTED) {}
	QuitCommand(DMessenger* msgr, const BMessage* specifier);
	virtual string Text() const;
};

class AboutCommand : public Command
{
public:
	AboutCommand(DMessenger* msgr) : Command(msgr, B_ABOUT_REQUESTED) {}
	AboutCommand(DMessenger* msgr, const BMessage* specifier);
	virtual string Text() const;
};

class SaveCommand : public Command
{
public:
	SaveCommand(DMessenger* msgr) : Command(msgr, B_SAVE_REQUESTED) {}
	SaveCommand(DMessenger* msgr, const BMessage* specifier);
	virtual string Text() const;
};

class LoadCommand : public Command
{
public:
	LoadCommand(DMessenger* msgr, const char* file);
	LoadCommand(DMessenger* msgr, const entry_ref* file);
	virtual status_t InitCheck() const;
	virtual string Text() const;

private:
	string path;
	status_t status;
	status_t AddFile(const char* file);
	status_t AddFile(const entry_ref* file);
};

class GetSuitesCommand : public Command
{
public:
	GetSuitesCommand(DMessenger* msgr) : Command(msgr, B_GET_SUPPORTED_SUITES) {}
	GetSuitesCommand(DMessenger* msgr, const BMessage* specifier);
	virtual status_t Do();
	virtual string Text() const;
};

class GetCommand : public Command
{
public:
	GetCommand(DMessenger* msgr, const BMessage* specifier);
	virtual status_t Do();
	virtual string Text() const;
};

class SetCommand : public Command
{
public:
	SetCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value);
	virtual string Text() const;
};

class DoCommand : public Command
{
public:
	DoCommand(DMessenger* msgr, const BMessage* specifier);
	DoCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value);
	virtual status_t Do();
	virtual string Text() const;
};

class DeleteCommand : public Command
{
public:
	DeleteCommand(DMessenger* msgr, const BMessage* specifier);
	virtual string Text() const;
};

class CountCommand : public Command
{
public:
	CountCommand(DMessenger* msgr, const BMessage* specifier);
	virtual status_t Do();
	virtual string Text() const;
};

class CreateCommand : public Command
{
public:
	CreateCommand(DMessenger* msgr, const BMessage* specifier);
	CreateCommand(DMessenger* msgr, const BMessage* specifier, const BMessage* value);
	virtual string Text() const;
};

class TellCommand : public Command
{
public:
	// Retains ownership of passed messenger and command list.
	TellCommand(DMessenger* msgr, string& target, CommandList* block);
	virtual ~TellCommand();
	virtual status_t Do();
	virtual void Print(int indent = 0);
	virtual void PrintContents(bool printReply = true);
	virtual string Text() const;

private:
	string		 fTarget;
	CommandList* fCommands;
};

class LetCommand : public Command
{
public:
	// Retains ownership of passed command list.
	LetCommand(DMessenger* msgr, BMessage* specifier, CommandList* block);
	virtual ~LetCommand();
	virtual status_t Do();
	virtual void Print(int indent = 0);
	virtual void PrintContents(bool printReply = true);
	virtual status_t InitCheck();
	virtual string Text() const;
	
private:
	CommandList* fCommands;
};

class LaunchCommand : public Command
{
public:
	LaunchCommand(DMessenger* msgr, const char* application);
	virtual status_t Do();
	virtual string Text() const;

private:
	string fApplication;
};

class TypeCommand : public Command
{
public:
	TypeCommand(DMessenger* msgr, BMessage* specifier, const BMessage* value);
	virtual string Text() const;
};

class ClickCommand : public Command
{
public:
	ClickCommand(DMessenger* msgr, BMessage* specifier);
	virtual string Text() const;
};

class VerbCommand : public Command
{
public:
	VerbCommand(DMessenger* msgr, uint32 verb) : Command(msgr, verb) {}
	VerbCommand(DMessenger* msgr, uint32 verb, BMessage* specifier);
	VerbCommand(DMessenger* msgr, uint32 verb, BMessage* specifier, BMessage* value);
	virtual string Text() const;
};

class ChooseCommand : public Command
{
public:
	ChooseCommand(DMessenger* msgr, BMessage* specifier);
	ChooseCommand(DMessenger* msgr, int32 item, BMessage* specifier);
	virtual string Text() const;
};

class PauseCommand : public Command
{
public:
	PauseCommand(DMessenger* msgr, int32 pause);
	virtual status_t Do();
	virtual string Text() const;

private:
	bigtime_t	fPause;
};

class PrintCommand : public Command
{
public:
	PrintCommand(DMessenger* msgr, const char* str);
	virtual status_t Do();
	virtual string Text() const;
	
private:
	string		fString;
};

// Currently equal to the TYPE command.
class AssignCommand : public Command
{
public:
	AssignCommand(DMessenger* msgr, BMessage* specifier, const BMessage* value);
	virtual string Text() const;
};

#endif	// ABSTRACTSYNTAX_H
