// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Application object.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

// System headers.

#include <cstdio>
#include <getopt.h>
#include <Application.h>
#include <Entry.h>

// Local headers.

#include "errormsg.h"
#include "yystype.h"
#include "parser.h"
#include "BeScriptApp.h"

// GLOBAL DATA:
// Verbosity level. See AbstractSyntax.cc for a description.

int BeScriptApp::globalVerbosity = 0;
int* gVerbosity = &BeScriptApp::globalVerbosity; 

// Application.

BeScriptApp::BeScriptApp(const char* signature)
: BApplication(signature), fArgvReceived(false), fPrintSyntax(false), fExecute(true)
{
	fError = false;
}

// We don't want to process a file from here -- we post a message to ourselves,
// and we process it from there. It's not nice to run things from ArgvReceived()
// because the message loop has not been executed yet -- only launched.
// There is a small penalty for running the compiler from the message loop, but
// it is better than having a misbehaving BApplication.

void BeScriptApp::ArgvReceived(int32 argc, char** argv)
{
	// Available options.
	struct option BeScriptLongOptions[] =
	{
		{ "syntax",        no_argument, NULL, 's' },
		{ "noexec",        no_argument, NULL, 'n' },
		{ "verbose",	   no_argument, NULL, 'v' },
		{ NULL,            no_argument, NULL, 0 }
	};

	// Prepare options scan.
	int c;
	int option_index = 0;
	opterr = 1;
	optind = 0;
	
	// Print usage information if we don't have anything on the command line.
	if (argc < 2)
	{
		PrintUsage(argv[0]);
		return;
	}
	
	// Scan for options.
	while (true)
	{
		c = getopt_long(argc, argv, "snv", BeScriptLongOptions, &option_index);
		if (c == -1)
			break;
		switch (c)
		{
			case 's': fPrintSyntax = true; break;
			case 'n': fExecute = false; break;
			case 'v': ++(*gVerbosity); break;
			default:  PrintUsage(argv[0]); return;
		}
	}
	
	// Make sure we still have valid parameters.
	if (optind >= argc)
	{
		PrintUsage(argv[0]);
		return;
	}
	
	// Now we have a valid attempt to launch the compiler.
	fArgvReceived = true;
	BMessage msg(BESCRIPT_APP_RUNSCRIPT);
	int32 total = argc - optind;
	msg.AddInt32("argc", total);
	for (int32 index = optind; index < argc; ++index)
		msg.AddString("argv", argv[index]);
	PostMessage(&msg);
}

// Dispatch running message.

void BeScriptApp::MessageReceived(BMessage* msg)
{
	switch (msg->what)
	{
		case BESCRIPT_APP_RUNSCRIPT:
		{
			// Make sure we have valid command line parameters.
			int32 argc;
			char** argv;
			status_t error = msg->FindInt32("argc", &argc);
			if (error != B_OK || argc < 1)
			{
				fprintf(stderr, "cannot find argc: %s\n", strerror(error));
				PostMessage(B_QUIT_REQUESTED);
				break;
			}
			
			// Extract parameters in a fresh local array.
			argv = new char* [argc];
			for (int32 index = 0; index < argc; ++index)
			{
				const char* argument;
				error = msg->FindString("argv", index, &argument);
				if (error != B_OK)
				{
					fprintf(stderr, "cannot find argv[%ld]: %s\n", index, strerror(error));
					PostMessage(B_QUIT_REQUESTED);
					return;
				}
				argv[index] = new char[strlen(argument) + 1];
				strcpy(argv[index], argument);
			}
			
			// Process files and dispose of command line parameters. Then, quit.
			ProcessFiles(argc, argv);
			for (int32 index = 0; index < argc; ++index)
				delete[] argv[index];
			delete[] argv;
			PostMessage(B_QUIT_REQUESTED);
		}
		break;
		default:
			BApplication::MessageReceived(msg);
	}
}

// If we reach this point without having processed anything, we must terminate.

void BeScriptApp::ReadyToRun()
{
	if (!fArgvReceived)
		PrintUsage("bs");
}

// Note that this expects only file arguments -- that is,
// argv[0] is *not* the application but rather the first script.

void BeScriptApp::ProcessFiles(int32 argc, char** argv)
{
	// Currently we refer to the global program from here.
	// Note that since yylex() and yyparse() are not thread-safe, we cannot structure
	// the compiler to be run concurrently.
	extern CommandList* gProgram;
	if (!gProgram)
		gProgram = new CommandList;
	for (int32 index = 0; index < argc; ++index)
	{
		if (parse(argv[index], fPrintSyntax))
		{
			if (fExecute)
			{
				if (*gVerbosity)
					printf("Starting: %s\n", argv[index]);
				status_t error = gProgram->Do();
				if (error != B_OK)
					return;
			}
			else
				fprintf(stderr, "%s successfully parsed\n", argv[index]);
		}
		else
			// Actually, yyerror() exits the program, so it's quite
			// difficult that the following code ever gets executed...
			// but these, as we all know, are the famous last words.
			break;
	}
	
	// Cleanup.
	gProgram->clear();
	delete gProgram;
	gProgram = 0;
}

// Display usage information.

void BeScriptApp::PrintUsage(const char* program)
{
	if (!fError)
	{
		fprintf(stderr,	"Usage: %s [OPTION ...] FILE [...]\n"
						"where: OPTION may be\n"
						"	-s, --syntax   print abstract syntax before running\n"
						"	-n, --noexec   parse only, do not execute\n"
						"	-v, --verbose  increment verbosity level (default=0,max=3)\n",
						program);
		fError = true;
	}
	PostMessage(B_QUIT_REQUESTED);
}
