//	Author:		Dario Accornero
//	BeDev ID:	10183
//	Mailto:		adario@cs.bu.edu
//	Project:	BeScript
//	Module:
//	A scripting element.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

//	System headers.

#ifndef _STDIO_H_
#include <stdio.h>
#endif
#ifndef	_WINDOW_H
#include <Window.h>
#endif

//	Local headers.

#include "ScriptingElement.h"
#include "ScriptingUtils.h"

//	CLASSES:
//	A ScriptingElement item is displayed in a BOutlineListView, so we need a class
//	suitable for inclusion in such list.

#ifndef BESCRIPT
//	This is similar to a BStringItem but it has two strings and they are drawn
//	using different colors.
SEItem::SEItem(const char* text, uint32 outlineLevel, bool expanded)
: BListItem(outlineLevel, expanded)
{
	//	Default text.
	SetText(text);
	SetDynamicText("");
	
	//	Needed colors.
	kHighlight = ui_color(B_PANEL_BACKGROUND_COLOR);
	rgb_color red = { 0xF0, 0x10, 0x00, 0xFF };
	kRedColor = red;
	kDimRedColor = tint_color(kRedColor, B_DISABLED_LABEL_TINT);
	rgb_color black = { 0x00, 0x00, 0x00, 0xFF };
	kBlackColor = black;
	kDimBlackColor = tint_color(kBlackColor, B_DISABLED_MARK_TINT);
}

void SEItem::DrawItem(BView* owner, BRect frame, bool complete)
{
	//	Keep copy of frame rectangle.
	if (fFrame != frame)
		fFrame = frame;
	
	//	Obtain item information needed to display the icon.
	BRect iconFrame(0, 0, 15, 15);
	iconFrame.OffsetTo(frame.left, frame.top);
	ScriptingElement* element = static_cast<ScriptingElement*>(this);
	const BBitmap* bitmap = element->GetBitmap();

	//	Retrieve font and compute information about the strings.
	//	Possibly truncate strings, currently in a brutal way.
	BString text(Text());
	BString dynamicText(DynamicText());
	float textLength = owner->StringWidth(text.String());
	float dynamicTextLength = owner->StringWidth(dynamicText.String());
	float width = frame.Width() - 16;
	float half = (width / 2) - 4;
	if (textLength + dynamicTextLength > width)
	{
		if (textLength > half)
		{
			owner->TruncateString(&text, B_TRUNCATE_SMART, half);
			textLength = owner->StringWidth(text.String());
		}
		if (dynamicTextLength > half)
		{
			owner->TruncateString(&dynamicText, B_TRUNCATE_SMART, half);
			dynamicTextLength = owner->StringWidth(dynamicText.String());
		}
	}
	
	//	Prepare background.
	rgb_color color;
	rgb_color highColor = owner->HighColor();
	rgb_color lowColor = owner->LowColor();
	if (IsSelected() || complete)
	{ 
		if (IsSelected())
		{
			color = kHighlight; 
			owner->SetLowColor(color);
		}
		else
			color = owner->ViewColor(); 
		owner->SetHighColor(color); 
		owner->FillRect(frame);
	}
	
	//	Compute text position and prepare pen.
	float x = frame.left + 4 + 16;
	float mid = (Height() - fHeight);
	float y = frame.bottom - mid;
	owner->MovePenTo(x, y);
	
	//	Draw bitmap.
	owner->PushState();
	owner->SetDrawingMode(B_OP_ALPHA);
	owner->SetBlendingMode(B_CONSTANT_ALPHA, B_ALPHA_OVERLAY);
	owner->DrawBitmap(bitmap, iconFrame);
	owner->PopState();
	
	//	Draw internal name.
	if (textLength > 0)
	{
		if (IsEnabled()) 
			owner->SetHighColor(kBlackColor); 
		else 
			owner->SetHighColor(kDimBlackColor); 
		owner->DrawString(text.String());
		owner->MovePenBy(4, 0);
	}

	//	Draw dynamic name.
	if (dynamicTextLength > 0)
	{
		if (IsEnabled()) 
			owner->SetHighColor(kRedColor);
		else 
			owner->SetHighColor(kDimRedColor); 
		owner->DrawString(dynamicText.String());
	}
	
	//	Restore drawing environment.
	owner->SetHighColor(highColor);
	owner->SetLowColor(lowColor);
}

//	Compute item dimensions.
void SEItem::Update(BView *owner, const BFont *font)
{
	//	Ensure item is at least as high as an icon plus space.
	const float minHeight = 16 + 2;
	BListItem::Update(owner, font);
	if (Height() < minHeight)
		SetHeight(minHeight);
	font_height height;
	font->GetHeight(&height);
	fHeight = height.ascent + height.descent + height.leading;
}

#endif	// !BESCRIPT

//	--- ScriptingElement:
//		Dynamic name resolution at analysis time.

bool ScriptingElement::DynamicNames = true;

//	--- ScriptingElement:
//		Constructor.

ScriptingElement::ScriptingElement(	BMessenger* theMessenger,
									ScriptingElement* theParent,
									const int32 theLevel )
#ifndef BESCRIPT
: SEItem("", theLevel, false)
#endif
{
	//	Assign default values and base members.
	//	Note that by default the object is *not* considered scriptable.
	
	SetMessenger(theMessenger);
	fParent = theParent;
	fChildren = 0;
	fLevel = theLevel;
	fIsScriptable = false;
	fInternalName = "";
	fName = "";
	fNumber = -1;
	fSuites = BMessage(B_MESSAGE_TYPE);
	fAccess = BMessage(B_MESSAGE_TYPE);
}

//	--- ScriptingElement:
//		Destructor.

ScriptingElement::~ScriptingElement()
{		
	//	Here we remove only the children and not the parent: it is a top-down traversal.
	//	Also, we don't remove the messenger since it might done at other levels.
	//	This will likely leave dangling pointers, but we're not ready for destruction yet.
	
	if (fChildren)
	{
		int32 items = fChildren->CountItems();
		for (int32 index = 0; index < items; ++index)
		{
			ScriptingElement* element = static_cast<ScriptingElement*>(fChildren->ItemAt(index));
			delete element;
		}
		fChildren->MakeEmpty();
		delete fChildren;
	}
}

//	--- ScriptingElement:
//		Display element on the console.

void ScriptingElement::PrintToStream() const
{
	//	By default, this method just calls the display function of all the children (if any).

	if (fChildren && fChildren->CountItems() > 0)
	{
		for (int32 index = 0; index < fChildren->CountItems(); ++index)
		{
			ScriptingElement* element = static_cast<ScriptingElement*>(fChildren->ItemAt(index));
			if (element)
				element->PrintToStream();
		}
	}
}

//	--- ScriptingElement:
//		Display default indentation on the console.

void ScriptingElement::IndentToStream() const
{
	for (int32 index = 0; index < fLevel; ++index)
		printf("    ");
}

//	--- ScriptingElement:
//		Retrieve supported suites.

status_t ScriptingElement::FindSuites()
{
	//	We ask the target messenger for its supported suites.
	//	Note that this is *not* done explicitly by ojects of this class:
	//	subclasses should call this method directly after initialization.
	//	ScriptingElement is an abstract class and cannot be instantiated directly.

	status_t result = B_ERROR;
	if (fMessenger && fMessenger->IsValid())
	{
		//	Be sure to use the base access message.
				
		BMessage getSupportedSuitesMsg(B_GET_SUPPORTED_SUITES);
		if (!fAccess.IsEmpty())
			AddAccess(&getSupportedSuitesMsg);
		result = fMessenger->SendMessage(&getSupportedSuitesMsg, &fSuites);
	}
	return	result;
}

//	--- ScriptingElement:
//		Retrieve element by name.

ScriptingElement* ScriptingElement::FindElement(const char* name)
{
	//	Check if this is what we're looking for. We look into both name fields.
	
	if (fInternalName == name || fName == name)
		return this;
		
	//	Now look into the children. It's a depth-first search -- should it be breadth-first? Dunno.
	
	if (fChildren && fChildren->CountItems() > 0)
	{
		int32 items = fChildren->CountItems();
		for (int32 index = 0; index < items; ++index)
		{
			ScriptingElement* element = static_cast<ScriptingElement*>(fChildren->ItemAt(index));
			if (element)
			{
				ScriptingElement* result = element->FindElement(name);
				if (result)
					return result;
			}
		}		
	}
	
	//	Out of luck -- no element with given name in current hierarchy.
	
	return 0;
}

//	--- ScriptingElement:
//		Add base access message to target message.

void ScriptingElement::AddAccess(BMessage* msg)
{
	//	Note that we cannot the whole base access message, since this would
	//	result in a nesting of specifiers, and the Be scripting system currently
	//	does not support this. Instead, we have to add one specifier at a time.
	
	int32 index = 0;
	BMessage specMsg;
	while (fAccess.FindMessage("specifiers", index++, &specMsg) == B_OK)
		msg->AddSpecifier(&specMsg);
}

//	--- ScriptingElement:
//		Retrieve dynamic name of element.

status_t ScriptingElement::FindDynamicName()
{
	//	By default we ask for a "Label".
	//	Obviously the element should support this property, otherwise this attempt is pointless.
	//	Unfortunately at the moment this crashes some Be-defined classes in certain occasions,
	//	including BTextControl and BSlider, when they are instantiated with a NULL label.
	
	BMessage reply(B_NO_REPLY);
	BMessage nameMsg(B_GET_PROPERTY);
	nameMsg.AddSpecifier(SM_LABEL);
	AddAccess(&nameMsg);
	status_t result = fMessenger->SendMessage(&nameMsg, &reply);
	if (result == B_OK)
	{
		const char* name;
		result = reply.FindString(SF_RESULT, &name);
		if (result == B_OK)
			if (fName != name)
				fName.SetTo(name);
	}
	#ifndef BESCRIPT
	else
	{
		#if DEBUG
		printf("Trying to get "); string str(specifier_to_str(&nameMsg)); printf("%s\n", str.c_str());
		printf("Object supports: "); print_message(&fSuites);
		printf("Message sent: "); print_message(&nameMsg);
		printf("Error: %s\n", strerror(result));
		#endif
	}
	#endif	// !BESCRIPT
	return result;
}

//	--- ScriptingElement:
//		Add entire hierarchy to list view. It is done using depth-first traversal.
#ifndef BESCRIPT
void ScriptingElement::AddToView(BOutlineListView* theView)
{
	BOutlineListView* view = dynamic_cast<BOutlineListView*>(theView);
	if (view && view->Window() && view->Window()->Lock())
	{
		view->AddItem(this);
		if (fChildren && fChildren->CountItems() > 0)
		{
			int32 items = fChildren->CountItems();
			for (int32 index = 0; index < items; index++)
			{
				ScriptingElement* child = static_cast<ScriptingElement*>(fChildren->ItemAt(index));
				if (child)
					child->AddToView(view);
			}
		}
		view->Window()->Unlock();
	}
}

void ScriptingElement::AddUnder(BOutlineListView* theView, ScriptingElement* parent)
{
	BOutlineListView* view = dynamic_cast<BOutlineListView*>(theView);
	if (view && view->Window() && view->Window()->Lock())
	{
		view->AddUnder(this, parent);
		if (fChildren && fChildren->CountItems() > 0)
		{
			int32 items = fChildren->CountItems();
			for (int32 index = items - 1; index >= 0; --index)
			{
				ScriptingElement* child = static_cast<ScriptingElement*>(fChildren->ItemAt(index));
				if (child)
					child->AddUnder(view, this);
			}
		}
		view->Window()->Unlock();
	}
}

//	--- ScriptingElement:
//		Reset names displayed in list and dynamic name as well.

void ScriptingElement::ResetNames()
{
	const char* empty = "";
	SetText(empty);
	SetDynamicText(empty);
	fName.SetTo(empty);
}
#endif	// !BESCRIPT

//	--- ScriptingElement:
//		Return name of parent (if any).

const char* ScriptingElement::ParentName()
{
	if (!fParent)
		return "(no parent)";
	else
	{
		if (fParent->fName.Length() > 0)
			return fParent->fName.String();
		else
			return fParent->fInternalName.String();
	}
}
