//	Author:		Dario Accornero
//	BeDev ID:	10183
//	Mailto:		adario@cs.bu.edu
//	Project:	BeScript
//	Module:
//	A scripting element.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

#ifndef	SCRIPTINGELEMENT_H
#define	SCRIPTINGELEMENT_H

//	System headers.

#ifndef _MESSENGER_H
#include <Messenger.h>
#endif
#ifndef	_BE_LIST_H
#include <List.h>
#endif
#ifndef	__BSTRING__
#include <String.h>
#endif
#ifndef BESCRIPT
#ifndef _LIST_ITEM_H
#include <ListItem.h>
#endif
#ifndef _OUTLINE_LIST_VIEW_H
#include <OutlineListView.h>
#endif
#ifndef PREFILLEDBITMAP_H
#include "PrefilledBitmap.h"
#endif
#endif	// !BESCRIPT

//	CONSTANTS:
//	Various scripting strings.
//	Currently defined as constants in the (childish) hope that the linker
//	will discard any string duplicate.

//	Be-defined fields.

#define	SF_SUITES 		"suites"
#define	SF_RESULT		"result"
#define	SF_ERROR		"error"
#define SF_SPECIFIERS	"specifiers"
#define SF_PROPERTY		"property"
#define SF_NAME			"name"
#define SF_INDEX		"index"
#define SF_RANGE		"range"
#define SF_MESSAGE		"message"
#define SF_MESSAGES		"messages"
#define SF_REFS			"refs"
#define SF_DATA			"data"
#define SF_ID			"id"

//	Be-defined properties.

#define SM_NAME			"Name"
#define	SM_INTERNALNAME	"InternalName"
#define	SM_WINDOW		"Window"
#define	SM_MESSENGER	"Messenger"
#define	SM_WINDOWS		"Windows"
#define	SM_VIEW			"View"
#define SM_MENU			"Menu"
#define	SM_MENUBAR		"MenuBar"
#define SM_MENUITEM		"MenuItem"
#define	SM_TITLE		"Title"
#define	SM_LABEL		"Label"
#define SM_VALUE		"Value"

//	CLASSES:
//	An item in the list to be displayed onscreen.

#ifndef BESCRIPT
class SEItem : public BListItem
{
public:
	// Basically two possible strings displayed in different colors and properly truncated.
	SEItem(const char* text, uint32 outlineLevel = 0, bool expanded = true);
	virtual	void DrawItem(BView* owner, BRect frame, bool complete = false);
	virtual void Update(BView *owner, const BFont *font);
	
	inline BRect Frame() const						{ return fFrame; }
	inline const char* Text() const					{ return fText.String(); }
	inline void SetText(const char* text)			{ fText.SetTo(text); }
	inline const char* DynamicText() const			{ return fDynamicText.String(); }
	inline void SetDynamicText(const char* text)	{ fDynamicText.SetTo(text); }

private:
	// The drawing code is blatantly taken from the Be Book.
	BRect		fFrame;
	BString		fText;
	BString		fDynamicText;
	rgb_color	kHighlight;
	rgb_color	kRedColor;
	rgb_color	kDimRedColor;
	rgb_color	kBlackColor;
	rgb_color	kDimBlackColor;
	float		fHeight;
};
#endif // !BESCRIPT

//	A scripting element may be a whole application, a window, or a view. In general,
//	any BHandler-derived object should be scriptable, but for the time being we only
//	target GUI objects.
//	All scriptable objects derive from this class, which cannot be instantiated directly
//	because it doesn't know how to analyze its children: it is an abstract class.

class _EXPORT ScriptingElement
#ifndef BESCRIPT
: public SEItem
#endif
{
public:													//	Single constructor (also works as default).
	ScriptingElement(	BMessenger* theMessenger = 0,
						ScriptingElement* theParent = 0,
						const int32 theLevel = 0 );
	virtual ~ScriptingElement();
	
	virtual void PrintToStream() const;					//	Display element on the console.
	virtual	void IndentToStream() const;				//	Display indentation on the console.
	virtual	int32 Analyze(int32 theLevel = 0) = 0;		//	Analyze element and its children.
	
	status_t FindSuites();								//	Retrieve supported suites.
	ScriptingElement* FindElement(const char* name);	//	Retrieve element by name.
	const char* ParentName();							//	Return name of parent (if any).
	
	virtual	void AddAccess(BMessage* msg);				//	Add base access message to target message.
	virtual status_t FindDynamicName();					//	Locate (if possible) element "screen name".
#ifndef BESCRIPT
	virtual void AddToView(BOutlineListView* theView);	//	Add this element and its children to a list view.
														//	Add this element under a given parent.
	virtual void AddUnder(BOutlineListView* theView, ScriptingElement* parent);
	virtual void ResetNames();							//	Reset names displayed in the list.
	virtual void UpdateItem() = 0;						//	Update item description.
	virtual const BBitmap* GetBitmap() const = 0;		//	Pure virtual to retrieve object bitmap.
#endif

	//	Accessors:
	//	Assigning information.
	
	inline void SetInternalName(const BString& theInternalName)	{ fInternalName.SetTo(theInternalName); }
	inline void SetSuites(BMessage* theSuites)					{ fSuites = *theSuites; }
	inline void SetLevel(const int32 theLevel)					{ fLevel = theLevel; }
	inline void SetMessenger(BMessenger* theMessenger)			{ fMessenger = theMessenger; }
	inline void SetMessenger(const BMessenger& theMessenger)
	{
		//	This variant is meant for situations when the element must retain messenger ownership.

		delete fMessenger;
		fMessenger = new BMessenger(theMessenger);
	}

	//	Retrieving information. Some methods are redundant but they're all inline...
	
	inline bool 			IsScriptable() const				{ return fIsScriptable; }
	inline const BString&	GetInternalName() const				{ return fInternalName; }
	inline const char* 		GetName() const						{ return fName.String(); }
	inline int32 			GetNumber() const					{ return fNumber; }
	inline const BMessage&	GetSuites() const					{ return fSuites; }	
	inline const BMessage*	GetSuitesPtr() const				{ return &fSuites; }
	inline const BMessage&	GetAccess() const					{ return fAccess; }
	inline const BMessage*	GetAccessPtr() const				{ return &fAccess; }
	inline BMessenger*		GetMessenger() const				{ return fMessenger; }
	inline int32			GetLevel() const					{ return fLevel; }

	//	Global settings.
	
	static bool			DynamicNames;		//	Retrieve dynamic name for element. Currently broken
											//	because of system bug #20001020-21887.

protected:									//	Members.

	bool				fIsScriptable;		//	Can we script this object?
	BString				fInternalName;		//	Internal name for this element.
	BString				fName;				//	"Real" (screen/dynamic) name for this element.
	int32				fNumber;			//	ID for this element in parent hierarchy.
	BMessage			fSuites;			//	Scripting suites supported by element.
	BMessage			fAccess;			//	Base specifier message to access this element.
	BMessenger*			fMessenger;			//	Target messenger for this element.
	int32				fLevel;				//	Element indentation level.
	BList*				fChildren;			//	Element children.
	ScriptingElement*	fParent;			//	Element parent.
};

#endif	//	SCRIPTINGELEMENT_H
