%{
// Author:		Dario Accornero <adario@cs.bu.edu>
// BeDev ID:	10183
// Project:		BeScript
// Module:
// Grammar specification.

/*
	Copyright 2000, Dario Accornero.   All Rights Reserved.
	This file may be used under the terms of the Modified Be Sample Code License
	(the "Software License" document accompanying this file).
*/

// System headers.

#ifndef __CSTDIO__
#include <cstdio>
#endif
#ifndef __SGI_STL_STACK
#include <stack>
#endif
#ifndef _ROSTER_H
#include <Roster.h>
#endif
#ifndef _APPLICATION_H
#include <Application.h>
#endif

// Local headers.

#include "errormsg.h"
#include "yystype.h"

// DEBUGGING:

#define YYDEBUG 0

// GLOBAL DECLARATIONS:
// Just types at the moment, may end up being unused.

typedef stack<CommandList*> CmdListStack;

// LOCAL DATA:
// Possible parser errors.

enum ParseErrors
{
	peNoError = 0,
	peUnknownExp,
	peUnknownTarget,
	peInvalidSpecifier,
	
	// Number of total errors for string representation.
	
	peNumErrors
};

// Strings for error messages.
// Currently unused.

static const char* parseErrors[peNumErrors] =
{
	"no error",
	"unknown expression type",
	"unknown target '%s'",
	"invalid specifier '%s'"
};

// Parsing data.
// Currently declared static which obviously makes it NOT thread-safe. However, yyparse() itself
// is not meant to be thread-safe, so there's not much we can do... short of turning on the
// multiple launch flag for the application. The final executable will be rather compact, so
// hopefully this won't be such a bad issue.

static CommandList*	pBlock = 0;
static Command*		pTell = 0;
static DMessenger*	pTarget = 0;
static BMessage		pSpecifier;
static BMessage		pData;
static string		pDataName;
static string		pTargetName;
static BMessage		pLetSpecifier;
static CmdListStack	pCmdListStack;

// EXTERNAL REFERENCES:
// The global command list for the whole program.

extern CommandList* gProgram;

// EXTERNAL FUNCTIONS:
// The lexer and the current text.

extern int			yylex(void);
extern char*		yytext;

// LOCAL FUNCTIONS:
// Standard error function.

void yyerror(const char* s);

// Conversions.

inline string i2str(int num)
{
	char buf[64];
	sprintf(buf, "%d", num);
	return string(buf);
}

inline string f2str(float num)
{
	char buf[64];
	sprintf(buf, "%.1f", num);
	return string(buf);
}

// Specifiers/data access.

inline void InitData(bool doInit = true)
{
	if (doInit)
	{
		pData = BMessage(B_ANY_TYPE);
		pDataName = "data";
	}
}

inline void InitSpecifier(bool initData = true)
{
	pSpecifier = BMessage(B_MESSAGE_TYPE);
	InitData(initData);
}

%}

/* Declarations for tokens that have an associated semantic value. */

%token <sval> ID STRINGID
%token <ival> INT
%token <fval> FP
%token <bval> BFALSE BTRUE
%token <wval> VERBID

/* Declarations for plain tokens. */
/* These represent punctuation, arithmetic/logical operators, and control keywords. */

%token 
	DOT COMMA COLON SEMICOLON LPAREN RPAREN LBRACK RBRACK LBRACE RBRACE
	PLUS MINUS TIMES DIVIDE EQ NEQ LE LT GE GT ASSIGNMENT AND OR
	IF THEN ELSE WHILE FOR TO DO LET IN END OF BREAK NIL
	SINGLEQUOTE

/* Declarations for tokens that describe the actual statements, */
/* raw specifiers interaction, and common Be data types. */

%token
	GETSUITES GET SET CREATE EXECUTE DELETE COUNT TELL ENDTELL QUIT ABOUT SAVE LOAD
	THE PROPERTY WITH RESULT VALUE MESSAGE BRECT BPOINT RGBCOLOR BFILE
	BOOL STRING INT8 INT16 INT32 FLOAT DOUBLE
	
/* Custom command tokens. */

%token LAUNCH CLICK BUTTON TYPE INTO FIELD CHOOSE PAUSE PRINT ASSIGN
	
/* Precedence/associativity for relevant operators. */
/* Currently unused. */

%nonassoc OR
%nonassoc AND
%nonassoc EQ NEQ GT LT GE LE
%left PLUS MINUS
%left TIMES DIVIDE
%left UMINUS
%left ASSIGNMENT

/* Semantic value for relevant nonterminals. */

%type <block> programme
%type <msgr> target preamble

%type <cmd>	statement statement_block _statement_block
			command command_block _command_block
			let_command_block _let_command_block

			tell_statement
			launch_statement

			quit_command
			about_command
			save_command
			load_command
			getsuites_command
			get_command
			set_command
			exec_command
			delete_command
			count_command
			create_command
			type_command
			click_command
			verb_command
			choose_command
			pause_command
			print_command
			assign_command

			let_command
			standard_let_command
			alternate_let_command
			inner_command

%type <sval> property_name

%type <ival> integer_value
%type <fval> fp_value
%type <sval> str_value string_value
%type <bval> truth_value bool_value
%type <int8val> inteight_value
%type <int16val> intsixteen_value
%type <int32val> intthirtytwo_value
%type <floatval> float_value
%type <doubleval> double_value
%type <pointval> point_value
%type <rectval> rect_value
%type <colorval> color_value
%type <fileval> file_value

/* Start symbol for the grammar. */

%start programme

%%

/* GRAMMAR PRODUCTIONS:
   A programme. */
/* Currently it is created in a global variable. Nothing is perfect. */

programme:			{	// Starting program analysis.
						// We store the code in a global variable, and we expect to reuse it
						// only sequentially -- not concurrently.
						if (gProgram)
							gProgram->clear();
						else
							gProgram = new CommandList;
						
						// This should be executed only once per programme,
						// so we're hopefully safe even though we'd actually
						// have to dispose of the contents as well as popping them.

						while (!pCmdListStack.empty())
							pCmdListStack.pop();
						
						// Turn on debugging if necessary.
						
						#if YYDEBUG != 0
						yydebug = 1;
						#endif
					}
					statement_block
					{ $$ = gProgram; }

/* STATEMENTS AND BLOCKS:
   All legal ones. */

statement_block:	statement
					{ gProgram->push_back($1); }
					_statement_block
					{
						if ($3)
							gProgram->push_back($3);
						$$ = $3;
					}

_statement_block:	statement_block
					{ $$ = $1; }
|
					{ $$ = 0; }

statement:			tell_statement
					{ $$ = $1; }
|					launch_statement
					{ $$ = $1; }
|					pause_command
					{ $$ = $1; }
|					print_command
					{ $$ = $1; }

tell_statement: 	preamble
					{	// Create a tell block that encompasses all of the following statements.
						pTarget = $1;
						pBlock = new CommandList;
						pTell = new TellCommand(pTarget, pTargetName, pBlock);
					}
					command_block postamble
					{ $$ = pTell; }
preamble:			TELL target
					{ $$ = $2; }
postamble:			ENDTELL
target:				ID
					{	// This is an application name.
						pTargetName = $1;
						$$ = new DMessenger($1.c_str());
					}
|					STRINGID
					{	// This is a MIME signature.
						pTargetName = $1;
						$$ = new DMessenger($1.c_str());
					}

launch_statement:	LAUNCH STRINGID
					{	// The MIME type must appear as a string -- quotes are needed.
						$$ = new LaunchCommand(new DMessenger(&be_app_messenger), $2.c_str());
						if ($$->InitCheck() != B_OK)
						{
							char errstr[256];
							sprintf(errstr, parseErrors[peUnknownTarget], $2.c_str());
							yyerror(errstr);
						}						
					}

/* Command blocks. These appear inside TELL blocks. */

command_block:		command
					{ pBlock->push_back($1); }
					_command_block
					{	// Return the last command parsed.
						if ($3)
							pBlock->push_back($3);
						$$ = $3;
					}
_command_block:		command_block
					{ $$ = $1; }
|
					{ $$ = 0; }
command:			quit_command
					{ $$ = $1; }
|					about_command
					{ $$ = $1; }
|					save_command
					{ $$ = $1; }
|					load_command
					{ $$ = $1; }
|					getsuites_command
					{ $$ = $1; }
|					get_command
					{ $$ = $1; }
|					set_command
					{ $$ = $1; }
|					exec_command
					{ $$ = $1; }
|					delete_command
					{ $$ = $1; }
|					count_command
					{ $$ = $1; }
|					create_command
					{ $$ = $1; }
|					type_command
					{ $$ = $1; }
|					click_command
					{ $$ = $1; }
|					verb_command
					{ $$ = $1; }
|					choose_command
					{ $$ = $1; }
|					pause_command
					{ $$ = $1; }
|					print_command
					{ $$ = $1; }
|					assign_command
					{ $$ = $1; }
|					let_command
					{ $$ = $1; }

/* The actual commands syntax. */

quit_command:		QUIT
					{ InitSpecifier(false); }
					specifier
					{ $$ = new QuitCommand(pTarget, &pSpecifier); }
|					QUIT
					{ $$ = new QuitCommand(pTarget); }
about_command:		ABOUT
					{ InitSpecifier(false); }
					specifier
					{ $$ = new AboutCommand(pTarget, &pSpecifier); }
|					ABOUT
					{ $$ = new AboutCommand(pTarget); }
save_command:		SAVE
					{ InitSpecifier(); }
					specifier
					{ $$ = new SaveCommand(pTarget, &pSpecifier); }
|					SAVE
					{ $$ = new SaveCommand(pTarget); }
load_command:		LOAD STRINGID
					{ $$ = new LoadCommand(pTarget, $2.c_str()); }
|					LOAD file_value
					{ $$ = new LoadCommand(pTarget, &$2); }
getsuites_command:	GETSUITES
					{ InitSpecifier(false); }
					of_preamble specifier
					{ $$ = new GetSuitesCommand(pTarget, &pSpecifier); }
|					GETSUITES
					{ $$ = new GetSuitesCommand(pTarget); }
of_preamble:		OF
|
get_command:		GET
					{ InitSpecifier(false); }
					specifier
					{ $$ = new GetCommand(pTarget, &pSpecifier); }
set_command:		SET
					{ InitSpecifier(); }
					specifier TO value
					{ $$ = new SetCommand(pTarget, &pSpecifier, &pData); }
exec_command:		exec_command_id
					{ InitSpecifier(); }
					specifier
					{	// Add data only if necessary.
						if (pData.IsEmpty())
							$$ = new DoCommand(pTarget, &pSpecifier);
						else
							$$ = new DoCommand(pTarget, &pSpecifier, &pData);
					}
exec_command_id:	DO
|					EXECUTE
delete_command:		DELETE
					{ InitSpecifier(false); }
					specifier
					{ $$ = new DeleteCommand(pTarget, &pSpecifier); }
count_command:		COUNT
					{ InitSpecifier(false); }
					specifier
					{ $$ = new CountCommand(pTarget, &pSpecifier); }
create_command:		CREATE
					{ InitSpecifier(); }
					specifier
					{	// Add data only if necessary.
						if (pData.IsEmpty())
							$$ = new CreateCommand(pTarget, &pSpecifier);
						else
							$$ = new CreateCommand(pTarget, &pSpecifier, &pData);
					}					
type_command:		TYPE str_value INTO field_preamble
					{	// Initiate specifier and data access.
						InitSpecifier();
						pData.AddString("data", $2.c_str());
					}
					specifier
					{ $$ = new TypeCommand(pTarget, &pSpecifier, &pData); }
field_preamble:		FIELD
|
click_command:		CLICK button_preamble
					{ InitSpecifier(false); }
					specifier
					{ $$ = new ClickCommand(pTarget, &pSpecifier); }
button_preamble:		BUTTON
|
verb_command:		DO VERBID
					{ InitSpecifier(); }
					specifier
					{ $$ = new VerbCommand(pTarget, $2, &pSpecifier); }
|					DO VERBID
					{ $$ = new VerbCommand(pTarget, $2); }
choose_command:		CHOOSE
					{ InitSpecifier(false); }
					specifier
					{ $$ = new ChooseCommand(pTarget, &pSpecifier); }
|					CHOOSE INT of_preamble
					{ InitSpecifier(false); }
					specifier
					{ $$ = new ChooseCommand(pTarget, $2, &pSpecifier); }
pause_command:		PAUSE INT
					{ $$ = new PauseCommand(pTarget, $2); }
print_command:		PRINT STRINGID
					{ $$ = new PrintCommand(pTarget, $2.c_str()); }
assign_command:		ASSIGN
					{ InitSpecifier(); }
					value TO specifier
					{ $$ = new AssignCommand(pTarget, &pSpecifier, &pData); }
let_command:		standard_let_command
					{ $$ = $1; }
|					alternate_let_command
					{ $$ = $1; }
standard_let_command:
					LET
					{ InitSpecifier(false); }
					specifier DO
					{	// Remember current specifier. Also, prepare a new block of commands.
						pLetSpecifier = pSpecifier;
						pCmdListStack.push(pBlock);
						pBlock = new CommandList;
					}
					let_command_block END
					{
						// Create let command and retrieve original command block.
						// We're not too much concerned about the target, because let commands
						// are not supposed to be nested right now...
						$$ = new LetCommand(pTarget, &pLetSpecifier, pBlock);
						pBlock = pCmdListStack.top();
						pCmdListStack.pop();
					}
alternate_let_command:
					TELL
					{ InitSpecifier(false); }
					specifier TO
					{	// Remember current specifier. Also, prepare a new block of commands.
						pLetSpecifier = pSpecifier;
						pCmdListStack.push(pBlock);
						pBlock = new CommandList;
					}
					inner_command
					{
						// Create let command and retrieve original command block.
						// We're not too much concerned about the target, because let commands
						// are not supposed to be nested right now...
						pBlock->push_back($6);
						$$ = new LetCommand(pTarget, &pLetSpecifier, pBlock);
						pBlock = pCmdListStack.top();
						pCmdListStack.pop();
					}
let_command_block:	inner_command
					{ pBlock->push_back($1); }
					_let_command_block
					{
						if ($3)
							pBlock->push_back($3);
						$$ = $3;
					}
_let_command_block:	let_command_block
					{ $$ = $1; }
|
					{ $$ = 0; }
inner_command:		quit_command
					{ $$ = $1; }
|					about_command
					{ $$ = $1; }
|					save_command
					{ $$ = $1; }
|					load_command
					{ $$ = $1; }
|					getsuites_command
					{ $$ = $1; }
|					get_command
					{ $$ = $1; }
|					set_command
					{ $$ = $1; }
|					exec_command
					{ $$ = $1; }
|					delete_command
					{ $$ = $1; }
|					count_command
					{ $$ = $1; }
|					create_command
					{ $$ = $1; }
|					type_command
					{ $$ = $1; }
|					click_command
					{ $$ = $1; }
|					verb_command
					{ $$ = $1; }
|					choose_command
					{ $$ = $1; }
|					pause_command
					{ $$ = $1; }
|					print_command
					{ $$ = $1; }
|					assign_command
					{ $$ = $1; }

/* SPECIFIERS:
   All available ones.
   A WITH block is considered part of a specifier but it is saved in a separate message. */

specifier:			spec_item _specifier
_specifier:			OF specifier
|					IN spec_item
					{	// Remember specifier to be resolved automatically.
						// Only one automatic specifier per complete specifier is allowed.
						int32 autoresolve = 0;
						pSpecifier.GetCurrentSpecifier(&autoresolve);
						int32 oldautoresolve;
						if (pSpecifier.FindInt32("autoresolve", &oldautoresolve) != B_OK)
							pSpecifier.AddInt32("autoresolve", autoresolve);
						else
							pSpecifier.ReplaceInt32("autoresolve", autoresolve);
					}
|					with_block
|

spec_item:			spec_preamble _spec_item
spec_preamble:		THE
|
/* The first rule supports constructs such as "do MoveBy BPoint(...)". */
_spec_item:			ID be_value
					{ pSpecifier.AddSpecifier($1.c_str()); }
|					property_name SINGLEQUOTE VERBID SINGLEQUOTE
					{
						BMessage spec($3);
						pSpecifier.AddSpecifier(&spec);
					}
|					property_name VERBID
					{
						BMessage spec(B_ID_SPECIFIER);
						spec.AddString("property", $1.c_str());
						spec.AddInt32("id", $2);
						pSpecifier.AddSpecifier(&spec);
					}
|					property_name STRINGID
					{ pSpecifier.AddSpecifier($1.c_str(), $2.c_str()); }
|					property_name INT
					{ pSpecifier.AddSpecifier($1.c_str(), int32($2)); }
|					property_name LBRACK INT RBRACK
					{ pSpecifier.AddSpecifier($1.c_str(), int32($3)); }
|					property_name MINUS INT
					{
						BMessage revspec(B_REVERSE_INDEX_SPECIFIER);
						revspec.AddString("property", $1.c_str());
						revspec.AddInt32("index", int32($3));
						pSpecifier.AddSpecifier(&revspec);
					}
|					property_name LBRACK MINUS INT RBRACK
					{
						BMessage revspec(B_REVERSE_INDEX_SPECIFIER);
						revspec.AddString("property", $1.c_str());
						revspec.AddInt32("index", int32($4));
						pSpecifier.AddSpecifier(&revspec);
					}
|					property_name LBRACK INT TO INT RBRACK
					{ pSpecifier.AddSpecifier($1.c_str(), int32($3), int32($5)); }
|					property_name LBRACK MINUS INT TO INT RBRACK
					{
						BMessage revspec(B_REVERSE_RANGE_SPECIFIER);
						revspec.AddString("property", $1.c_str());
						revspec.AddInt32("index", int32($4));
						revspec.AddInt32("range", int32($6));
						pSpecifier.AddSpecifier(&revspec);
					}
|					property_name
					{ pSpecifier.AddSpecifier($1.c_str()); }

property_name:		ID
					{ $$ = $1; }
|					STRINGID
					{ $$ = $1; }

value:				integer_value
					{ pData.AddInt32(pDataName.c_str(), $1); }
|					fp_value
					{ pData.AddFloat(pDataName.c_str(), $1); }
|					str_value
					{ pData.AddString(pDataName.c_str(), $1.c_str()); }
|					truth_value
					{ pData.AddBool(pDataName.c_str(), $1); }
|					bool_value
					{ pData.AddBool(pDataName.c_str(), $1); }
|					string_value
					{ pData.AddString(pDataName.c_str(), $1.c_str()); }
|					inteight_value
					{ pData.AddInt8(pDataName.c_str(), $1); }
|					intsixteen_value
					{ pData.AddInt16(pDataName.c_str(), $1); }
|					intthirtytwo_value
					{ pData.AddInt32(pDataName.c_str(), $1); }
|					float_value
					{ pData.AddFloat(pDataName.c_str(), $1); }
|					double_value
					{ pData.AddDouble(pDataName.c_str(), $1); }
|					point_value
					{ pData.AddPoint(pDataName.c_str(), $1); }
|					rect_value
					{ pData.AddRect(pDataName.c_str(), $1); }
|					color_value
					{ pData.AddData(pDataName.c_str(), B_RGB_COLOR_TYPE, &$1, sizeof(rgb_color)); }
|					file_value
					{	// We make sure to add a corresponding entry for B_REFS_RECEIVED.
						pData.AddRef(pDataName.c_str(), &$1);
						if (pDataName != "refs")
							pData.AddRef(pDataName.c_str(), &$1);
					}
/* This is a kludge needed for custom commands with direct invokation and single value. */
be_value:			bool_value
					{ pData.AddBool(pDataName.c_str(), $1); }
|					string_value
					{ pData.AddString(pDataName.c_str(), $1.c_str()); }
|					inteight_value
					{ pData.AddInt8(pDataName.c_str(), $1); }
|					intsixteen_value
					{ pData.AddInt16(pDataName.c_str(), $1); }
|					intthirtytwo_value
					{ pData.AddInt32(pDataName.c_str(), $1); }
|					float_value
					{ pData.AddFloat(pDataName.c_str(), $1); }
|					double_value
					{ pData.AddDouble(pDataName.c_str(), $1); }
|					point_value
					{ pData.AddPoint(pDataName.c_str(), $1); }
|					rect_value
					{ pData.AddRect(pDataName.c_str(), $1); }
|					color_value
					{ pData.AddData(pDataName.c_str(), B_RGB_COLOR_TYPE, &$1, sizeof(rgb_color)); }
|					file_value
					{	// We make sure to add a corresponding entry for B_REFS_RECEIVED.
						pData.AddRef(pDataName.c_str(), &$1);
						if (pDataName != "refs")
							pData.AddRef(pDataName.c_str(), &$1);
					}

/* Numeric constants management is quite rough at the moment. */

integer_value:		MINUS INT
					{ $$ = -$2; }
|					INT
					{ $$ = $1; }
fp_value:			MINUS FP
					{ $$ = -$2; }
|					FP
					{ $$ = $1; }
str_value:			ID
					{ $$ = $1; }
|					STRINGID
					{ $$ = $1; }
truth_value:		BFALSE
					{ $$ = false; }
|					BTRUE
					{ $$ = true; }
bool_value:			BOOL LPAREN truth_value RPAREN
					{ $$ = $3; }
string_value:		STRING LPAREN STRINGID RPAREN
					{ $$ = $3; }
inteight_value:		INT8 LPAREN integer_value RPAREN
					{ $$ = int8($3); }
intsixteen_value:	INT16 LPAREN integer_value RPAREN
					{ $$ = int16($3); }
intthirtytwo_value:	INT32 LPAREN integer_value RPAREN
					{ $$ = int32($3); }
float_value:		FLOAT LPAREN fp_value RPAREN
					{ $$ = float($3); }
double_value:		DOUBLE LPAREN fp_value RPAREN
					{ $$ = double($3); }
point_value:		BPOINT LPAREN fp_value COMMA fp_value RPAREN
					{ $$ = BPoint($3, $5); }
rect_value:			BRECT LPAREN fp_value COMMA fp_value COMMA fp_value COMMA fp_value RPAREN
					{ $$ = BRect($3, $5, $7, $9); }
color_value:		RGBCOLOR LPAREN integer_value COMMA integer_value COMMA integer_value COMMA integer_value RPAREN
					{
						rgb_color color = { $3, $5, $7, $9 };
						$$ = color;
					}
file_value:			BFILE LPAREN str_value RPAREN
					{
						BEntry entry($3.c_str());
						entry.GetRef(&$$);
					}

/* WITH data blocks. */
/* The syntax is WITH [name=]value [AND [name=]value ...]. */
/* Its weirdness is reflected in two harmless shift/reduce conflicts at bison time, */
/* and also in a somewhat rare initial mid-rule for a non-initial production. */

with_block:			WITH
					{ InitData(); }
					_with_block
_with_block:		data_item __with_block
__with_block:		AND _with_block
|
data_item:			ID ASSIGNMENT
					{ pDataName = $1; }
					value
|					{ pDataName = "data"; }
					value

/* EXPRESSIONS:
   All legal expression types. */
/* Currently unused. */

%%

// Additional parsing functions:
// Standard error call.

void yyerror(const char* s)
{
	EM_error(EM_tokPos, "\n%s", s);
	EM_close();
	exit(1);
}
